#!/usr/bin/python2.4
#    gcode "remaining material" previewer
#    Copyright (C) 2007 Jeff Epler <jepler@unpythonic.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import Image, numarray, itertools, sys, rs274.interpret, gcode, getopt, os
import numarray.ufunc
import numarray.ieeespecial as ieee
from math import *

def ball_tool(r,dia):
    #print r, dia,
    s = -sqrt((dia/2)**2-r**2)
    #print s
    return s
def endmill(r,dia):
    return 0

def vee_common(angle):
    slope = tan(angle * pi / 180)
    print "vee_common", angle, slope
    def f(r, dia):
        return r * slope
    return f
    
tool_makers = {
    'b': ball_tool, 'e': endmill, 'V': vee_common(45), 'v': vee_common(30)}


def make_tool_shape(f, wdia):
    dia = int(wdia*res+.5)
    if dia < 2: dia = 2
    n = numarray.array([[ieee.plus_inf] * dia] * dia, type="Float32")
    hdia = dia / 2.
    l = []
    for x in range(dia):
        for y in range(dia):
            r = hypot(x-hdia, y-hdia)
            if r < hdia:
                #print r,
                z = f(r*resp, wdia)
                l.append(z)
                n[x,y] = z
    print "Tool:", min(l), max(l)
    n = n - n.min()
    print n.min(), n.max()
    #n = n * dia
    return n

def cut_at(x, y, z):
    #print x, y, z
    x = int((x-minx) * res + .5 - ts0/2)
    y = int((y-miny) * res + .5 - ts0/2)

    if x < 0 or y < 0: return
    #print tool.shape, a[x:x+ts,y:y+ts].shape

    try:
	a[x:x+ts,y:y+ts] = numarray.ufunc.minimum(a[x:x+ts,y:y+ts], z + tool)
    except ValueError:
	pass # too close to the edge

def cut_along(x1, y1, z1, x2, y2, z2):
    dx = x2-x1
    dy = y2-y1
    dz = z2-z1
    if dx == 0 and dy == 0:
	cut_at(x1, y1, min(z1, z2))
        return 
    dt = 1./max(abs(dx), abs(dy)) / res

    #print x1, y1, z1, "...", x2, y2, z2
    for i in itertools.count():
	t = i * dt
	if t > 1: break
	cut_at(x1 + t*dx, y1 + t*dy, z1 + t*dz)

def triangle(x1,y1,x2,y2,x3,y3):
    z1 = a[x1,y1]
    z2 = a[x2,y2]
    z3 = a[x3,y3]
    x1 *= resp
    x2 *= resp
    x3 *= resp
    y1 *= resp
    y2 *= resp
    y3 *= resp
    dx1 = x2-x1
    dx2 = x3-x1
    dy1 = y2-y1
    dy2 = y3-y1
    dz1 = z2-z1
    dz2 = z3-z1

    nx = dy1 * dz2 - dz1 * dy2
    ny = dz1 * dx2 - dx1 * dz2
    nz = dx1 * dy2 - dy1 * dx2

    glNormal3f(nx, ny, nz)
    glVertex3f(x3+minx,y3+miny,z3)
    glVertex3f(x2+minx,y2+miny,z2)
    glVertex3f(x1+minx,y1+miny,z1)

def triangles():
    glBegin(GL_TRIANGLES)
    for x in range(shape[0]-1):
        for y in range(shape[1]-1):
            triangle(x,y, x,y+1, x+1,y)
            triangle(x,y+1, x+1,y+1, x+1,y)
        sys.stderr.write("."); sys.stderr.flush()
    glEnd()
    sys.stderr.write("\n"); sys.stderr.flush()

try:
    import _gdepth
except ImportError, detail:
    print >>sys.stderr, "_gdepth speed-up module not available: %s" % detail
else:
    def cut_along(x1, y1, z1, x2, y2, z2):
        x1 = int((x1-minx) * res + .5 - ts0 / 2)
        y1 = int((y1-miny) * res + .5 - ts0/2)
        x2 = int((x2-minx) * res + .5 - ts0 / 2)
        y2 = int((y2-miny) * res + .5 - ts0/2)
        _gdepth.putline(x1, y1, z1, x2, y2, z2, a, tool)
    def triangles():
        glPushMatrix();
        glTranslatef(minx, miny, 0);
        _gdepth.triangles(res, a);
        glPopMatrix();

class Canon(rs274.interpret.Translated, rs274.interpret.ArcsToSegmentsMixin):
    def __init__(self):
	self.lo = (0,0,0)

    def straight_traverse(self, x, y, z, a, b, c):
	cut_along(self.lo[0], self.lo[1], self.lo[2], x, y, z)
	self.lo = x, y, z
    straight_feed = straight_traverse

    def comment(self, comment):
        print comment
        parts = comment.split(",")
        if len(parts) < 2: return
        if parts[0].lower() == "tool":
            global maker, tool, tool_size, ts
            if len(parts) == 3:
                maker = tool_makers.get(parts[1], endmill)
                tool_size = float(parts[2])
            else:
                tool_size = float(parts[1])
            new_tool = make_tool_shape(maker, tool_size)
	    tool[:] = ieee.plus_inf
	    ts1 = new_tool.shape[0]
            tl = (ts-ts1)/2
	    tool[tl:tl+ts1,tl:tl+ts1] = new_tool
            #ts = tool.shape[0]
            print "new tool", maker, tool_size, ts
            #print tool
    def set_feed_rate(self, rate): pass
    def dwell(self, arg): pass
    def next_line(self, arg): pass
    def check_abort(self): pass

    def get_external_length_units(self): return 1.0
    def get_external_angular_units(self): return 1.0

    def get_tool(self, tool):
	print "get_tool", tool
	return tool,0.,0.,0.,0.,0.,0

    def tool_offset(self, x, y): pass

    parameter_file = "ignore.var"

res = 128
minx, maxx = 0, 7.5
miny, maxy = -.1, 1
defz = 0
defzmin = -1
tool_size = (1/8.)
maker = ball_tool
out = None
display = True

opts, args = getopt.getopt(sys.argv[1:], "t:r:x:y:z:o:bvVe")
for k, v in opts:
    if k == "-V": maker = vee_common(60)
    elif k == "-v": maker = vee_common(45)
    elif k == "-b": maker = ball_tool
    elif k == "-e": maker = endmill
    elif k == "-t": tool_size = float(v)
    elif k == "-r": res = int(v)
    elif k == "-x":
	if len(v.split()) == 2:
	    minx, maxx = map(float, v.split())
	else:
	    maxx = float(v)
    elif k == "-y":
	if len(v.split()) == 2:
	    miny, maxy = map(float, v.split())
	else:
	    maxy = float(v)
    elif k == "-z":
	if len(v.split()) == 2:
	    defzmin, defz = map(float, v.split())
	else:
	    defz = float(v)
    elif k == "-o": out = v
    elif k == "-d": display = not display
if out is None: out = os.path.splitext(args[0])[0] + ".png"
resp = 1./res
 
sx = int((maxx - minx) * res)
sy = int((maxy - miny) * res)
cx =(maxx + minx)/2
cy =(maxy + miny)/2
tool = make_tool_shape(maker, tool_size)
#print "Tool:", tool.min(), tool.max()
ts0 = ts = tool.shape[0]
a = numarray.zeros((sx+ts, sy+ts), "Float32")
a[:] = defzmin
a[ts/2:-ts/2,ts/2:-ts/2] = defz

code, seq = gcode.parse(args[0], Canon(), 'g20 g17 g64', '')
if code > gcode.MIN_ERROR:
	raise SystemExit, "%s:%d: %s" % (args[0], seq, gcode.strerror(code))
top = a.max()
bottom = a.min()
#print top, bottom

if top == bottom: raise SystemExit, "No detail in resulting depth map"
if top != bottom:
    ai = (a - bottom) * 254 / (top-bottom)
else:
    ai = a[:,:]
#print a.max(), a.min()
ai = ai.astype('Int8')
shape = a.shape
ai = ai[:,::-1]
ai.transpose()
i = Image.fromstring("L", shape, ai.tostring())
i.save(out)
print >>sys.stderr, "Wrote", out

if not display: raise SystemExit, 0

import Tkinter, rs274.OpenGLTk
from minigl import *
app = Tkinter.Tk()

class Opengl(rs274.OpenGLTk.Opengl):
    def __init__(self, *args, **kw):
        self.l = None
        rs274.OpenGLTk.Opengl.__init__(self, *args, **kw)
        self.r_back = self.g_back = self.b_back = 0
    def basic_lighting(self):
        self.activate()
        glClearColor(0,0,0,0)
        glLightfv(GL_LIGHT0, GL_POSITION, (1, -1, -.5, 0))
        glLightfv(GL_LIGHT0, GL_AMBIENT, (.2,.2,.2,0))
        glLightfv(GL_LIGHT0, GL_DIFFUSE, (.6,.6,.4,0))
        glLightfv(GL_LIGHT0+1, GL_POSITION, (-1, -1, -.5, 0))
        glLightfv(GL_LIGHT0+1, GL_AMBIENT, (.0,.0,.0,0))
        glLightfv(GL_LIGHT0+1, GL_DIFFUSE, (.0,.0,.4,0))
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, (1,1,1,0))
        glEnable(GL_LIGHTING)
        glEnable(GL_LIGHT0)
        glEnable(GL_LIGHT0+1)
        glDepthFunc(GL_LESS)
        glEnable(GL_DEPTH_TEST)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

    def redraw(self, *args):
        if self.winfo_width() == 1: return
        if self.l is None:
            self.basic_lighting()
            glClearColor(0,0,0,0)
            self.l = make_list()
        glCallList(self.l)

def make_list():
    l = glGenLists(1)
    glEnable(GL_CULL_FACE)
    glEnable(GL_NORMALIZE)
    glNewList(l, GL_COMPILE)
    glEnable(GL_LIGHTING)
    glDisable(GL_POLYGON_OFFSET_FILL)
    glMatrixMode(GL_MODELVIEW)
    glPushMatrix()
    glTranslatef(-cx, -cy, 0)
    glColor3f(1,1,1)
    triangles()
    glNormal3f(0,0,-1)
    glVertex3f(minx,miny,defzmin)
    glVertex3f(minx,maxy+ts*resp,defzmin)
    glVertex3f(maxx+ts*resp,miny,defzmin)
    glVertex3f(minx,maxy+ts*resp,defzmin)
    glVertex3f(maxx+ts*resp,maxy+ts*resp,defzmin)
    glVertex3f(maxx+ts*resp,miny,defzmin)
    glDisable(GL_LIGHTING)
    glEnable(GL_POLYGON_OFFSET_FILL)
    glPolygonOffset(1,1)
    glBegin(GL_TRIANGLES)
    glColor3f(1,0,0)
    tso = (ts0 - 1) * resp
    glVertex3f(maxx+tso,miny,defzmin)
    glVertex3f(maxx+tso,maxy+tso,defzmin)
    glVertex3f(minx,maxy+tso,defzmin)
    glVertex3f(maxx+tso,miny,defzmin)
    glVertex3f(minx,maxy+tso,defzmin)
    glVertex3f(minx,miny,defzmin)
    glEnd()
    glPopMatrix()
    glEndList()
    return l

print a.min(), a.max()

t = Opengl(app, depth=1, double=1)
t.pack(fill="both", expand=1)
t.activate()
app.mainloop()

# vim:sts=4:et:sw=4
